//============================================================================
//
//    Java Text Editor library
//
//    HighLightedDocument.java
//
//  Copyright (c) 2013 Fumiaki Hosaka.
//
//  Permission to copy, use, modify, sell and distribute this software
//  is granted provided this copyright notice appears in all copies.
//  This software is provided "as is" without express or implied
//  warranty, and with no claim as to its suitability for any purpose.
//============================================================================
package text.editor;

import java.awt.*;
import java.awt.event.*;
import java.util.*;

import javax.swing.*;
import javax.swing.event.*;
import javax.swing.text.*;
import javax.swing.undo.*;

import java.util.concurrent.locks.*;

public class HighLightedDocument extends DefaultStyledDocument
{
    static class MargeAttCell
    {
       public int    offset;
       public int    len;
       public Style  att;
    }

    private ActionListener loadEndAction = null;

    private Lock gard = new ReentrantLock();

    public boolean isAttributeChange = false;

    SourceToolBar      tool    = null;
    JTextPane          text    = null;

    ArrayList<int []> backup  = null;

    HighLightControler highLightCntrl = null;

    Object lock = new Object();

    String srctext = null;

    Style  a_space    = null;
    Style  a_enter    = null;
    Style  a_comment  = null;
    Style  a_const    = null;
    Style  a_regular  = null;
    Style  a_class    = null;

    boolean isVisible = false;

    public boolean isThreadRunning = false;

    public HighLightedDocument(SourceToolBar tool, JTextPane text)
    {
        super();
        this.tool  = tool;
        this.text  = text;
    }

    public void dispose()
    {
        loadEndAction = null;
        gard    = null;
        tool    = null;
        text    = null;

        if (backup != null)
        {
            backup.clear();
            backup = null;
        }
        highLightCntrl = null;
        lock = null;
        srctext = null;
        a_space    = null;
        a_enter    = null;
        a_comment  = null;
        a_const    = null;
        a_regular  = null;
        a_class    = null;
    }

    protected void finalize() throws Throwable 
    {
        try
        {
            super.finalize();
        }
        finally
        {
            dispose();
        }
    }

    public void setLoadEndAction(ActionListener act)
    {
        this.loadEndAction = act;
    }

    public void loadEnded()
    {
        if (loadEndAction != null)
        {
            SwingUtilities.invokeLater(new Runnable()
            {
                public void run()
                {
                    loadEndAction.actionPerformed(null);
                }
            });
        }
    }

    protected void createStylesToDocument(Object tbl)
    {
        try
        {
            highLightCntrl = new HighLightControler(this, text, tbl);
            highLightCntrl.setStyles();
        }
        catch(Exception ex)
        {
            ex.printStackTrace();
            highLightCntrl = null;
        }

        if (highLightCntrl == null)
        {
           System.out.println("HighLight controle off!!");
        }
        else
        {
            getDefaultStyles();
           //highLightCntrl.dump();
        }
    }

    protected void resetStylesToDocument()
    {
        if (highLightCntrl == null)
        {
           System.out.println("resetStyles is null");
            return;
        }

        try
        {
            highLightCntrl.resetStyles();

            getDefaultStyles();
        }
        catch(Exception ex)
        {
            ex.printStackTrace();
            highLightCntrl = null;
        }
    }

    private void getDefaultStyles()
    {
        a_space   = null;
        a_enter   = null;
        isVisible = false;

        a_comment = getStyle("comment");
        a_const   = getStyle("constant");
        a_regular = getStyle("regular");
        a_class   = getStyle("_DEFINECLASS_");
    }

    public void insertString(int offs, String str, AttributeSet a) throws BadLocationException
    {
        try
        {
            gard.lock();

            tool.undoManager.push(new TextUndoCommand(TextUndoCommand.DELETE, offs, str.length()));
            if (highLightCntrl == null)
            {
                super.insertString(offs,str, a);
                return;
            }

            if (backup == null)
            {
                insertStringAddAtt(offs, str, null);
                srctext = tool.delCR(text.getText());
            }
            else
            {
                insertStringToken(offs, str, a_regular);
                srctext = tool.delCR(text.getText());

                partUpdateAtt(offs, str.length());
            }
        }
        finally
        {
           gard.unlock();
        }
    }

    public void insertStringSuper(int offs, String str, AttributeSet a) throws BadLocationException
    {
        super.insertString(offs, str, a);
    }

    private ArrayList<MargeAttCell> margeAttribute(ArrayList<int []> tok,
                                                   int start, int end,
                                                   String txt)
    {
        ArrayList<MargeAttCell> out = new ArrayList<MargeAttCell>();

        Style back = null;
        int offset_s = -1;
        int cnt      = 0;
        int [] t = null;
        int tsize = tok.size();
        for (int i = start; i < end && i < tsize; ++i)
        {
             t = tok.get(i);
             Style st = getUserStyle(t, txt);

             if( back == null
             || (st == back
             && (!isVisible || (isVisible && st != a_space &&  st != a_enter)) ))
             {
                 if (offset_s == -1)
                     offset_s = t[0];

                 cnt++;
             }
             else
             {
                 MargeAttCell c = new MargeAttCell();
                 c.offset = offset_s;
                 c.len    = t[0] - offset_s;
                 c.att    = back;
                 out.add(c);

                 offset_s = t[0];
                 cnt = 0;
             }
             back = st;
         }

         if (t != null && t[0] - offset_s != 0)
         {
             MargeAttCell c = new MargeAttCell();
             c.offset = offset_s;
             c.len    = t[0] - offset_s;
             c.att    = (back == null) ? a_regular : back;
             out.add(c);
         }
         return out;
    }

    class InsertStrThread implements Runnable
    {
        int offs;
        ArrayList<MargeAttCell> mgcell;

        public InsertStrThread(int offs, ArrayList<MargeAttCell> mgcell)
        {
            this.offs   = offs;
            this.mgcell = mgcell;
        }
        public void run()
        {
            try
            {
                isThreadRunning = true;
                for (final MargeAttCell m : mgcell)
                {
                    SwingUtilities.invokeLater(new Runnable()
                    {
                        public void run()
                        {
                            try
                            {
                                // TODO :
                                insertStringSuper(offs + m.offset,
                                                  srctext.substring(m.offset, m.len - m.offset -1),
                                                  m.att);
                            }
                            catch(BadLocationException ex) {  }
                        }
                    });
                }
            }
            catch(Exception ex) {  }

            text.setEnabled(true);
            loadEnded();
            isThreadRunning = false;
        }
    }

    public void insertStringToken(int offs, String str, AttributeSet a) throws BadLocationException
    {
        if (a_enter == null && a_space == null)
        {
            super.insertString(offs, str, a);
            return;
        }

        int add = 0;
        StringTokenizer tok = new StringTokenizer(str, "\r\n", true);
        while (tok.hasMoreTokens())
        {
            String s = tok.nextToken();
            char c = s.charAt(0);
            if (c == '\n')
            {
                super.insertString(offs + add, "\n", a_enter);
            }
            else
            {
                super.insertString(offs + add, s, a);
            }
            add += s.length();
        }
    }

    public void insertStringAddAtt(int offs, String str, AttributeSet a)
    {
        Lexer lex = new Lexer(str, 0, isVisible);
        lex.setLexicalChars(highLightCntrl);

        backup = lex.getTokens();
        lex.dispose();
        lex = null;

        srctext = tool.delCR(str);

        int startpos = 500;
        int div      = 10;

        ArrayList<MargeAttCell> mgcell = margeAttribute(backup, 0, startpos, srctext);

        try
        {
            for (MargeAttCell m : mgcell)
            {
                 super.insertString(offs + m.offset,
                                    str.substring(m.offset, m.len - m.offset -1),
                                   m.att);
            }
        }
        catch(Exception e) {}

        if (backup.size() < startpos)
           return;

        mgcell = margeAttribute(backup, startpos, backup.size(), srctext);
        Runnable th2 = new InsertStrThread(offs, mgcell);
        LayoutQueue.getDefaultQueue().addTask(th2);
    }

    public void remove(int offs, int len) throws BadLocationException
    {
        try
        {
            gard.lock();

            String s =  tool.delCR(text.getText());
            tool.undoManager.push(new TextUndoCommand(
                                TextUndoCommand.INSERT, offs, s.substring(offs, offs + len), null));

            if (highLightCntrl == null)
            {
                super.remove(offs, len);
                return;
            }

            super.remove(offs, len);

            srctext = tool.delCR(text.getText());

            if (backup == null)
            {
                allUpdateAtt();
            }
            else
            {
                partUpdateAtt(offs, -len);
            }
        }
        finally
        {
            gard.unlock();
        }
    }

   class AttupdateThread implements Runnable
   {
        ArrayList<MargeAttCell> mgcell;

        public AttupdateThread(ArrayList<MargeAttCell> mgcell)
        {
            this.mgcell = mgcell;
        }

        public void run()
        {
            isThreadRunning = true;
            try
            {
                for (final MargeAttCell m : mgcell)
                {
                    SwingUtilities.invokeLater(new Runnable()
                    {
                        public void run()
                        {
                            setCharacterAttributes(m.offset, m.len,  m.att, true);
                        }
                    });
                }
            }
            catch(Exception ex) {  }
            text.setEnabled(true);
            loadEnded();
            isThreadRunning = false;
        }
    }

    public void allUpdateAttThread(ArrayList<int []> w)
    {
        if (w.size() == 0)
            return;

        int startpos = 500;
        int div      = 10;

        try
        {
            ArrayList<MargeAttCell> mgcell = margeAttribute(w, 0, startpos, srctext);
            for (final MargeAttCell m : mgcell)
            {
                SwingUtilities.invokeLater(new Runnable()
                {
                    public void run()
                    {
                        setCharacterAttributes(m.offset, m.len,  m.att, true);
                    }
                });
            }

            if (w.size() < startpos)
            {
                loadEnded();
                return;
            }

            mgcell = margeAttribute(w, startpos, w.size(), srctext);
            Runnable th2 = new AttupdateThread(mgcell);
            LayoutQueue.getDefaultQueue().addTask(th2);
        }
        catch(Exception ex)
        {
        }
    }

    public void allUpdateAtt()
    {
        Lexer lex = new Lexer(srctext, 0, isVisible);
        lex.setLexicalChars(highLightCntrl);

        backup = lex.getTokens();

        lex.dispose();
        lex = null;

        allUpdateAttThread(backup);
    }

    public void refreshAtt()
    {
        srctext = tool.delCR(text.getText());
        Lexer lex = new Lexer(srctext, 0, isVisible);
        lex.setLexicalChars(highLightCntrl);

        backup = lex.getTokens();

        lex.dispose();
        lex = null;

        allUpdateAttThread(backup);
    }

    private int getLexPos(int offs)
    {
        int end = getLength() - 1;

        if (offs > end)
            offs = end;

        if (offs < 0)
           return  -1;

        if (srctext.charAt(offs) == '/')
        {
            if (offs  > 0 && srctext.charAt(offs -1) == '*')
            {
                for (int w = offs - 2; w >= 0 ; w--)
                {
                    if (srctext.charAt(w) == '/' && srctext.charAt(w + 1) == '*')
                    {
                        int wpos = seekToken(w);
                        if (wpos < 0)
                            return -1;

                        int [] wtok = (int [] )backup.get(wpos);
                        if (wtok[2] != Lexer.TOK_CONST && wtok[2] != Lexer.TOK_COMMENT_M
                                                        && wtok[2] != Lexer.TOK_COMMENT_S)
                           return w;
                    }
                }
            }
        }
        else if (srctext.charAt(offs) == '*')
        {
            if (offs < end && srctext.charAt(offs + 1) == '/')
            {
                for (int w = offs - 1; w >= 0 ; w--)
                {
                    if (srctext.charAt(w) == '/' && srctext.charAt(w + 1) == '*')
                    {
                        int wpos = seekToken(w);
                        if (wpos < 0)
                            return -1;

                        int [] wtok = (int [] )backup.get(wpos);
                        if (wtok[2] != Lexer.TOK_CONST && wtok[2] != Lexer.TOK_COMMENT_M
                                                        && wtok[2] != Lexer.TOK_COMMENT_S)
                           return w;
                    }
                }
            }
        }
        else if (srctext.charAt(offs) == '\"')
        {
            for (int w = offs - 1; w >=0 && srctext.charAt(w) != '\n' ; w--)
            {
                if (srctext.charAt(w) == '\"'
                    && (w == 0 || (w > 0 && srctext.charAt(w - 1) != '\\') ))
                {
                    if (w == 0)
                       return w;

                    int wpos = seekToken(w);
                    if (wpos < 0)
                        return -1;

                    if ((backup.get(wpos))[2] != Lexer.TOK_CONST)
                       return w;
                }
            }
        }
        return -1;
    }

    private void partUpdateAtt(int offs, int len)
    {
        try
        {
            gard.lock();

            int lexpos = getLexPos(offs);

            int pos = 0;
            if (lexpos >= 0)
            {
                pos = seekToken(lexpos);
            }
            else
            {
                pos = seekToken(offs);
                lexpos = 0;
                if (pos  >= 0)
                    lexpos = (backup.get(pos))[0];
            }

            ArrayList<int []> newtok = new ArrayList<int []>();
            ArrayList<int []> updt   = new ArrayList<int []>();

            Lexer lex = new Lexer(srctext, lexpos, isVisible);
            lex.setLexicalChars(highLightCntrl);

            int same = -1;

            for (int i = 0; i < pos; ++i)
                newtok.add(backup.get(i));

            while(true)
            {
               int [] tok = lex.getToken();
               if (tok == null)
                   break;

               same = seekSameToken(tok, len, (pos < 0) ? 0 : pos + 1);
               if (same > 0)
               {
                   updt.add(tok);
                   break;
               }

               newtok.add(tok);
               updt.add(tok);
            }

            lex.dispose();
            lex = null;

            if (same >= 0)
            {
                for (int i = same; i < backup.size(); ++i)
                {
                    int [] tok = backup.get(i);

                    tok[0] += len;
                    tok[1] += len;
                    newtok.add(tok);
                }
            }

            backup = newtok;

            if (updt.size() != 0);
            {
                ArrayList<MargeAttCell> mgcell = margeAttribute(updt, 0, updt.size(), srctext);
                for (MargeAttCell m : mgcell)
                {
                    setCharacterAttributes(m.offset, m.len,  m.att, true);
                }
            }
        }
        finally
        {
            gard.unlock();
        }
    }

    private void dump()
    {
        for (int i = 0; i < backup.size(); ++i)
        {
            int [] tok = backup.get(i);
            dump(tok, "[" + i + "] ");
        }
    }

    private void dump(int [] tok, String h)
    {
        String s = srctext.substring(tok[0],tok[1]).replace("\n", "\\n");
        String t = s.replace("\t", "\\t");
        String r = t.replace("\r", "\\r");

        System.out.println(h + tok[0] + ":" + tok[1] + ":" + tok[2] + ":\"" + r + "\"");
    }

    private int seekToken(int offs)
    {
        int max = backup.size();
        for (int i = 0; i < max; ++i)
        {
            int [] tok  = backup.get(i);
            if (offs >= tok[0] && offs <= tok[1])
                return i;
        }
        return  -1;
    }

    private int seekSameToken(int [] tok, int slen, int start)
    {
        int i   = start;
        int end = backup.size();

        for (;  i < end; ++i)
        {
           int [] old = backup.get(i);
           if (old[2] == tok[2]  && (old[0] + slen) == tok[0]
                                 && (old[1] + slen) == tok[1])
               return i;
        }
        return -1;
    }

    private void dumpElement(Element elem)
    {
        AbstractDocument.AbstractElement ae =  (AbstractDocument.AbstractElement)elem;
        ae.dump(System.out, 0);
    }

    public Style getUserStyle(int [] tok, String ctxt)
    {
         Style st = null;

         if (tok[2] == Lexer.TOK_COMMENT_S
          || tok[2] == Lexer.TOK_COMMENT_M)
         {
             st = a_comment;
         }
         else if (tok[2] == Lexer.TOK_CONST)
         {
             st = a_const;
         }
         else if (tok[2] == Lexer.TOK_NUMBER)
         {
             st = a_regular;
         }
         else
         {
             String key = ctxt.substring(tok[0], tok[1]);
             st = getStyle(key);
         }

         if (st == null)
             st = a_regular;

         return st;
    }

    public void initText(String s)
    {
        try
        {
            gard.lock();

            super.remove(0, getLength());

            insertStringToken(0, s, a_regular);

            tool.undoManager.push(new TextUndoCommand(TextUndoCommand.DELETE, 0, s.length()));

            srctext = tool.delCR(text.getText());
            allUpdateAtt();
        }
        catch(Exception e) {}
        finally
        {
            gard.unlock();
        }
    }

    public void textInit()
    {
        try
        {
            gard.lock();
            remove(0, getLength());
            backup = null;
        }
        catch(Exception e) {}
        finally
        {
            gard.unlock();
        }
    }

    public void fireUpdate()
    {
         resetStylesToDocument();
         
         DefaultDocumentEvent ev = new DefaultDocumentEvent(0, getLength(),
                                                            DocumentEvent.EventType.CHANGE);

         super.fireChangedUpdate(ev);
    }
}

