//============================================================================
//
//    Java Text Editor library
//
//    Lexer.java
//
//  Copyright (c) 2013 Fumiaki Hosaka.
//
//  Permission to copy, use, modify, sell and distribute this software
//  is granted provided this copyright notice appears in all copies.
//  This software is provided "as is" without express or implied
//  warranty, and with no claim as to its suitability for any purpose.
//============================================================================
package text.editor;

import java.util.*;

public class Lexer
{
    final static int TOK_END       = -1;
    final static int TOK_SPACE     = 0;
    final static int TOK_LF        = 1;
    final static int TOK_TAB       = 2;
    final static int TOK_GRAPH     = 3;
    final static int TOK_COMMENT_S = 4;
    final static int TOK_COMMENT_M = 5;
    final static int TOK_IDENT     = 6;
    final static int TOK_NUMBER    = 7;
    final static int TOK_CONST     = 8;
    final static int TOK_EOF       = 100;

    final static char EOF    = 0;    // sentinel \0
    final static char LCHAR  = '\'';
    final static char LSTR   = '\"';
    final static char LSLASH = '/';
    final static char LASTER = '*';
    final static char LSEMI  = ';';
    final static char ESC    = '\\';
    final static char LDOT   = '.';
    final static char LF     = '\n';
    final static char CR     = '\r';
    final static char TAB    = '\t';
    final static char SPC    = ' ';
    final static char ZSPC   = '　';

    String numStartChar = "0123456789.+-";
    String numNextChar  = "0123456789.";

    String idStartChar = "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ_$#";
    String idNextChar  = "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789_$#";
    String idEndChar   = "";
    String cnstChar    = "\"\'";
    String graphKeyChar = "";
    String onelineComChar = "";

    int comType = 0; // 0:C , 1:xml  2:oneline 

    char [] src  = null;
    int     pos  = 0;
    int     len  = 0;

    ArrayList<int []> tokens = null;
    ArrayList<int[]>  bktokens = new ArrayList<int[]>();

    String text = null;

    public boolean isVisible = false;

    public Lexer(String src, int offs)
    {
        this(src, offs, true);
    }

    public Lexer(String src, int offs, boolean isVisible)
    {
        this.text = src;
        this.src = (src + "        ").toCharArray(); // sentinel 8byte
        this.len = src.length();
        this.tokens = new ArrayList<int []>();
        this.pos = offs;
        this.src[len] = EOF; // sentinel
        this.isVisible = isVisible;
    }

    public void setLexicalChars(HighLightControler cntrl)
    {
        idStartChar = cntrl.idStartChar;
        idNextChar  = cntrl.idNextChar;
        idEndChar   = cntrl.idEndChar;
        cnstChar    = cntrl.cnstChar;
        graphKeyChar = cntrl.graphKeyChar;
        comType = cntrl.comType;
        onelineComChar = cntrl.onelineComChar;
    }

    public void dispose()
    {
        tokens = null;
        src = null;
    }

    protected void finalize() throws Throwable 
    {
        try
        {
            super.finalize();
        }
        finally
        {
            dispose();
        }
    }

    public ArrayList<int []>  getTokens()
    {
        int [] tok;
        while((tok = getToken()) != null)
              tokens.add(tok);

        return tokens;
    }

    public int[] getToken()
    {
        while(true)
        {
            if (src[pos] == EOF)
                return null;

            int start_pos = pos;
            char c = src[pos++];

            if (c == TAB)
                return new int [] { start_pos, pos, TOK_TAB };
            else if (c == LF)
                return new int [] { start_pos, pos, TOK_LF };
            else if (c == CR)
                continue;
            // white space
            else if (c == SPC)
            {
                if (!isVisible)
                {
                    while ( (c = src[pos]) == SPC )
                         pos++;
                }
                return new int [] { start_pos, pos, TOK_SPACE };
            }
            // identifier
            else if (idStartChar.indexOf(c) != -1)
            {
                char o = c;
                while (true)
                {
                    c = src[pos];
                    if (idNextChar.indexOf(c) == -1)
                        break;

                    pos++;
                    if (idEndChar.length()!= 0 && idEndChar.indexOf(c) != -1)
                        break;
                }

                if ((pos - start_pos) == 1 
                &&  graphKeyChar.indexOf(o) == -1)
                    return new int [] { start_pos, pos, TOK_GRAPH };

                if (comType == 2
                 && onelineComChar.length() > 0 && onelineComChar.charAt(0) == src[start_pos])
                {
                    String ss = new String(src,start_pos, pos - start_pos);
                    if (ss.equals(onelineComChar))
                    {
                        src[len] = LF;
                        while ((c = src[pos]) != LF)
                            pos++;

                        src[len] = EOF;
                        return new int [] { start_pos, pos, TOK_COMMENT_S };
                    }
                }
                return new int [] { start_pos, pos, TOK_IDENT };
            }
            // number
            else if (numStartChar.indexOf(c) != -1 )
            {
                char o = c;
                boolean dotstart = (c == '.');
                while (true)
                {
                    c = src[pos];
                    if (dotstart && c == '.')
                        break;
                    if (numNextChar.indexOf(c) == -1)
                        break;
                    pos++;
                }

                if ((pos - start_pos) == 1
                &&  graphKeyChar.indexOf(o) == -1)
                    return new int [] { start_pos, pos, TOK_GRAPH };

                return new int [] { start_pos, pos, TOK_IDENT};
            }
            // comment
            // C,C++/Java/CS
            else if (c == LSLASH && comType == 0)
            {
                c = src[pos];
                if (c == LSLASH)
                {
                    pos++;
                    src[len] = LF;
                    while ((c = src[pos]) != LF)
                        pos++;

                    src[len] = EOF;
                    return new int [] { start_pos, pos, TOK_COMMENT_S };
                }
                else if (c == LASTER)
                {
                    int bk = pos;
                    pos++;
                    src[len] = LASTER;
                    src[len + 1] = LSLASH;
                    while (!((c = src[pos]) == LASTER && src[pos + 1] == LSLASH))
                    {
                        if (c == LSLASH && pos < len - 1 && src[pos + 1] == LASTER)
                            return new int [] { start_pos, pos, TOK_COMMENT_M};
                        pos++;
                    }

                     src[len] = EOF;
                     if (pos == len)
                     {
                         pos = bk;
                         return new int [] { start_pos, pos, TOK_GRAPH };
                    }
                    else
                    {
                         pos += 2;
                        return new int [] { start_pos, pos, TOK_COMMENT_M };
                    }
                }
                else
                {
                    if (graphKeyChar.indexOf(c) != -1)
                        return new int [] { start_pos, pos, TOK_IDENT };
                    else
                        return new int [] { start_pos, pos, TOK_GRAPH };
                }
            }
            // MarkUp(XML/HTML)
            else if (c == '<' && comType == 1)
            {
                int bkt = pos;
                c = src[pos];
                pos++;

                if (c == '!' && pos < len - 1
                && src[pos] == '-' && src[pos + 1] == '-')
                {
                    int bk = pos;
                    pos++;
                    src[len] = '-';
                    src[len + 1] = '-';
                    src[len + 2] = '>';
                    while (!((c = src[pos]) == '-'
                             && src[pos + 1] == '-' && src[pos + 2] == '>'))
                    {
                        if (c == '<' && pos < len - 1
                        && src[pos + 1] == '!'
                        && src[pos + 2] == '-' && src[pos + 3] == '-')
                            return new int [] { start_pos, pos, TOK_COMMENT_M };

                        pos++;
                    }

                    src[len] = EOF;
                    if (pos == len)
                    {
                        pos = bk;
                        if (graphKeyChar.indexOf(c) != -1)
                            return new int [] { start_pos, pos, TOK_IDENT };
                        else
                            return new int [] { start_pos, pos, TOK_GRAPH };
                    }
                    else
                    {
                        pos += 4;
                        return new int [] { start_pos, pos, TOK_COMMENT_M };
                    }
                }
                else
                {
                    pos = bkt; 
                    if (graphKeyChar.indexOf(c) != -1)
                        return new int [] { start_pos, pos, TOK_IDENT };
                    else
                        return new int [] { start_pos, pos, TOK_GRAPH };
                }
            }
            // Oneline (lisp/Batch)
            else if (comType == 2 
                && onelineComChar.length() > 0 && onelineComChar.charAt(0) == c)
            {
                int ll= onelineComChar.length();
                boolean match = true;
                for (int i = 1; i < ll; ++i )
                {
                    if (i < len - 1)
                    {
                        match = false;
                        break;
                    }
                    if (src[i] != onelineComChar.charAt(i))
                    {
                        match = false;
                        break;
                    }
                }
                if (!match)
                {
                    if (graphKeyChar.indexOf(c) != -1)
                        return new int [] { start_pos, pos, TOK_IDENT };
                    else
                        return new int [] { start_pos, pos, TOK_GRAPH };
                }

                src[len] = LF;
                while ((c = src[pos]) != LF)
                    pos++;

                src[len] = EOF;
                return  new int [] { start_pos, pos, TOK_COMMENT_S };
            }
            // constant
            else if (cnstChar.indexOf(c) != -1)
            {
                if (src[pos] == EOF)
                {
                    if (graphKeyChar.indexOf(c) != -1)
                        return new int [] { start_pos, pos, TOK_IDENT };
                    else
                        return new int [] { start_pos, pos, TOK_GRAPH };
                }
                else
                {
                    int bk = pos;
                    src[len] = ' ';
                    src[len + 1] = c;
                    int escflg = -1;
                    while ( !(src[pos] == c && escflg == -1))
                    {
                        if (src[pos] == LF)
                        {
                            src[len] = EOF;
                            return new int [] { start_pos, pos, TOK_CONST };
                        }
                        if (src[pos] == ESC)
                        {
                            escflg *= -1;
                        }
                        else
                        {
                            if (escflg == 1)
                                escflg = -1;
                        }
                        pos++;
                    }

                    src[len] = EOF;

                    if (pos >= len + 1)
                    {
                        pos = bk;
                        return new int [] { start_pos, pos, TOK_IDENT };
                    }
                    else
                    {
                        pos++;
                        return new int [] { start_pos, pos, TOK_CONST };
                    }
                }
            }
            // GRAPH
            else
            {
                if (graphKeyChar.indexOf(c) != -1)
                    return new int [] { start_pos, pos, TOK_IDENT };
                else
                    return new int [] { start_pos, pos, TOK_GRAPH };
            }
        }
    }
}
