//============================================================================
//
//    Java Text Editor library
//
//    FontDialog.java
//
//  Copyright (c) 2013 Fumiaki Hosaka.
//
//  Permission to copy, use, modify, sell and distribute this software
//  is granted provided this copyright notice appears in all copies.
//  This software is provided "as is" without express or implied
//  warranty, and with no claim as to its suitability for any purpose.
//============================================================================
package text.editor;

import java.awt.*;
import java.awt.event.*;
import java.awt.geom.*;

import javax.swing.*;
import javax.swing.border.*;

public class FontDialog extends JDialog implements ActionListener
{
    private String [] fontList  = null;
    private String [] sizeList  = {"8", "10", "12", "14", "16", "18", "24", "32", "48", "64" };
    private JComboBox name   = null;
    private JComboBox size   = null;
    private JCheckBox bold   = null;
    private JCheckBox italic = null;
    private SamplePanel sample = null;

    private JButton ok = null;
    private JButton ng = null;
    private String result;

    public FontDialog(Frame owner, Font f)
    {
        super(owner, "Font", true);

        Box vbox = Box.createVerticalBox();
        vbox.add(createPane());
        vbox.add(createBtns());
        getContentPane().add(vbox, BorderLayout.CENTER);

        setListener();

        setSize(300, 200);

        Dimension size = Toolkit.getDefaultToolkit().getScreenSize();
        setLocation(size.width /2 - getSize().width /2,
                    size.height/2 - getSize().height/2);

        if (f != null)
            setInitSelection(f);

        pack();
    }

    private JComponent createPane()
    {
        JPanel p = new JPanel();
        GridBagConstraints gbc = new GridBagConstraints();
        GridBagLayout gbl = new GridBagLayout();
        p.setLayout(gbl);

        fontList =
                GraphicsEnvironment.getLocalGraphicsEnvironment()
                                   .getAvailableFontFamilyNames();

        name   = new JComboBox(fontList);
        size   = new JComboBox(sizeList);
        bold   = new JCheckBox("Bold");
        italic = new JCheckBox("Italic");
        sample = new SamplePanel();

        sample.setPreferredSize(new Dimension(300, 80));

        gbc.anchor = GridBagConstraints.WEST;
        gbc.insets = new Insets(5,5,5,5);
        gbc.gridx  = 1;       gbc.gridy  = 1;
        p.add(new JLabel("フォント名"), gbc);

        gbc.gridx  = 2;       gbc.gridy  = 1;
        gbc.gridwidth = 2;
        p.add(name, gbc);

        gbc.gridx  = 1;       gbc.gridy  = 2;
        gbc.gridwidth = 1;
        p.add(new JLabel("サイズ"), gbc);

        gbc.gridx  = 2;       gbc.gridy  = 2;
        size.setPreferredSize(new Dimension(50, 20));
        p.add(size, gbc);

        gbc.gridx  = 1;       gbc.gridy  = 3;
        gbc.gridwidth = 1;
        p.add(new JLabel("スタイル"), gbc);

        gbc.gridx  = 2;       gbc.gridy  = 3;
        p.add(bold, gbc);

        gbc.gridx  = 3;       gbc.gridy  = 3;
        p.add(italic, gbc);

        gbc.fill = GridBagConstraints.HORIZONTAL;
        gbc.gridwidth = 3;
        gbc.gridheight= 2;
        gbc.gridx  = 1;       gbc.gridy  = 4;
        p.add(sample, gbc);

        return p;
    }

    private JComponent createBtns()
    {
        JPanel p = new JPanel();

        ok = new JButton("Ok");
        ng = new JButton("Cancel");

        p.add(Box.createHorizontalGlue());
        p.add(ok);
        p.add(ng);
        p.add(Box.createHorizontalGlue());

        return p;
    }

    private void setListener()
    {
        name.addActionListener(this);
        size.addActionListener(this);
        bold.addActionListener(this);
        italic.addActionListener(this);

        ActionListener act =  new ActionListener()
        {
            public void actionPerformed(ActionEvent e)
            {
                ending(e);
            }
        };

        ok.addActionListener(act);
        ng.addActionListener(act);
    }

    private void setInitSelection(Font f)
    {
        String n = f.getName();
        name.setSelectedItem(n);

        String s = "" + f.getSize();
        size.setSelectedItem(s);

        bold.setSelected(f.isBold());
        italic.setSelected(f.isItalic());
    }

    public void ending(ActionEvent e)
    {
        JButton s = (JButton)e.getSource();
        result = null;

        if (s.getText().equals("Ok"))
            setResult();

        dispose();
    }

    public void actionPerformed(ActionEvent e)
    {
        sample.repaint();
    }

    private Font createFont()
    {
        String n = (String) name.getSelectedItem();

        int fstyle = 0;
        if (bold.isSelected())
            fstyle  = Font.BOLD;
        else
            fstyle  = Font.PLAIN;

        if (italic.isSelected())
            fstyle  += Font.ITALIC;

        String s = (String) size.getSelectedItem();

        int fsize = 12;
        try
        {
           fsize = Integer.parseInt(s);
        }
        catch( Exception ex) { }

        return new Font(n, fstyle, fsize);
    }

    private void setResult()
    {
        String fname  = (String) name.getSelectedItem();

        String fstyle = "";
        if (bold.isSelected() && italic.isSelected())
            fstyle  = "BOLD_ITALIC";
        else  if (italic.isSelected())
            fstyle  = "ITALIC";
        else  if (bold.isSelected())
            fstyle  = "BOLD";
        else
            fstyle  = "PLAIN";

        String fsize = (String) size.getSelectedItem();

        result = fname + ":" + fstyle + ":" + fsize;
    }

    public String getResult()
    {
         return result;
    }

    public Font getResultFont()
    {
         return (result == null)? null : createFont();
    }

    public class SamplePanel extends JPanel
    {
        private String drawStr = "Sample text";

        public SamplePanel()
        {
        }

        public void setDrawString(String str)
        {
            if(str == null
            || str.equals(""))
                return;

            drawStr = str;
        }

        public void paint(Graphics g)
        {
            Font f = createFont();

            g.setFont(f);
            g.setColor(Color.white);
            g.fillRect(0, 0, getWidth(), getHeight());

            g.setColor(Color.gray);
            g.drawRect(0, 0, getWidth(), getHeight());

            g.setColor(Color.black);
            g.getFontMetrics();

            Rectangle2D rect = g.getFontMetrics().getStringBounds(drawStr, g);

            int locateX = getWidth()/2  - g.getFontMetrics().stringWidth(drawStr)/2;
            int locateY = getHeight()/2 + (int)(rect.getHeight()/4);

            g.drawString(drawStr, getCorrectValue(locateX), getCorrectValue(locateY));
        }

        private int getCorrectValue(int value)
        {
            if(value < 0) return 0;

            return value;
        }
    }

}
