//============================================================================
//
//    Java Text Editor library
//
//    SourceToolBar.java
//
//  Copyright (c) 2013 Fumiaki Hosaka.
//
//  Permission to copy, use, modify, sell and distribute this software
//  is granted provided this copyright notice appears in all copies.
//  This software is provided "as is" without express or implied
//  warranty, and with no claim as to its suitability for any purpose.
//============================================================================
package text.editor;

import java.lang.reflect.*;
import java.awt.*;
import java.awt.event.*;
import java.util.*;
import java.util.concurrent.*;
import java.io.*;

import javax.swing.*;
import javax.swing.border.*;
import javax.swing.event.*;
import javax.swing.text.*;
import javax.swing.undo.*;
import java.beans.PropertyChangeListener;

import static text.editor.Kit.*;

public class SourceToolBar extends JToolBar
{
    static final Hashtable<JEditorPane, SourceToolBar> toolTbl
                           = new Hashtable<JEditorPane, SourceToolBar>();

    private JButton cutBtn     = null;
    private JButton copyBtn    = null;
    private JButton pasteBtn   = null;
    private JButton undoBtn    = null;
    private JButton redoBtn    = null;
    private JButton searchBtn  = null;
    private JButton macroRecBtn = null;
    private JButton macroPlayBtn = null;
    private JButton settingBtn = null;

    private JLabel     searchLabel   = null;
    private JTextField incSearchText = null;
    private JLabel     lineLabel   = null;
    private JTextField lineText    = null;
    private JLabel     colLabel    = null;
    private JTextField colText     = null;

    JEditorPane text = null;
    private EditorToolsDialog    toolsDialog =  null;
    static private EditSettingDialog  settingDlg  =  null;

    public TextUndoManager undoManager = null;
    private UndoAction  undoAction  = new UndoAction();
    private RedoAction  redoAction  = new RedoAction();

    private SourceFocusListener srcFocusListener =  new SourceFocusListener();
    private ActionListener codeParseAction = null;
    private Document doc = null;
    private LNView lnview = null;
    private KeyBindings keybinds = null;

    /////////////////////////////////////////////////////////////////////////////

    public SourceToolBar(JTextPane text, String ext)
    {
        super();
        this.text = text;

        Object tbl = HighLightControler.getExtEntry(ext);

        if (tbl != null)
        {
            HighLightedDocument hdoc = new HighLightedDocument(this, text);
            doc = hdoc;
            hdoc.createStylesToDocument(tbl);
        }
        else
        {
            NormalDocument ndoc = new NormalDocument(this, text);
            doc = ndoc;
        }
        text.putClientProperty(JEditorPane.HONOR_DISPLAY_PROPERTIES, Boolean.TRUE);
        text.setDocument(doc);
        init();
    }

    public void setLineNumberView(LNView lnview)
    {
        this.lnview = lnview;
    }

    private void init()
    {
        undoManager = new TextUndoManager(text, doc, undoChangeListener);

        text.setForeground(EditorConfig.fgColor);
        text.setBackground(EditorConfig.bgColor);

        keyBinding();

        createItem();
        createDialog();
        setListener();

        text.setCaretPosition(0);

        toolTbl.put(text, this);
    }

    public void hidePoolBtn()
    {
    }

    public void hideReadOnly()
    {
        cutBtn.setVisible(false);
        pasteBtn.setVisible(false);
        undoBtn.setVisible(false);
        redoBtn.setVisible(false);

        settingBtn.setVisible(false);
    }

    private void createItem()
    {
        cutBtn     = new JButton (getIconImage("Cut.GIF"));
        cutBtn.setToolTipText("切り取り");

        copyBtn    = new JButton (getIconImage("Copy.GIF"));
        copyBtn.setToolTipText("コピー");

        pasteBtn   = new JButton (getIconImage("Paste.GIF"));
        pasteBtn.setToolTipText("貼り付け");

        undoBtn   = new JButton (getIconImage("Undo.GIF"));
        undoBtn.setToolTipText("アンドゥ");
        undoBtn.setEnabled(false);

        redoBtn   = new JButton (getIconImage("Redo.GIF"));
        redoBtn.setToolTipText("リドゥ");
        redoBtn.setEnabled(false);

        searchBtn  = new JButton (getIconImage("Search.GIF"));
        searchBtn.setToolTipText("検索/置換");

        macroRecBtn   = new JButton (getIconImage("MacroOff.GIF"));
        macroRecBtn.setToolTipText("入力を記録");

        macroPlayBtn   = new JButton (getIconImage("MacroPlay.GIF"));
        macroPlayBtn.setToolTipText("記録を再生");

        settingBtn   = new JButton (getIconImage("Setting.GIF"));
        settingBtn.setToolTipText("エディタの設定");

        add(cutBtn);
        add(copyBtn);
        add(pasteBtn);

        addSeparator();
        add(undoBtn);
        add(redoBtn);

        addSeparator();
        add(searchBtn);

        addSeparator();
        add(macroRecBtn);
        add(macroPlayBtn);

        addSeparator();
        add(settingBtn);

        add(Box.createHorizontalGlue());

        JPanel p  = new JPanel();
        HistoryText h1 = new  HistoryText(32);
        h1.loadHistory("TextEditor.incSearch");
        incSearchText = h1.text;
        incSearchText.setColumns(12);
        incSearchText.setToolTipText("インクリメンタルサーチ.");
        lineLabel = new JLabel("line");
        lineText = new JTextField(4);
        colLabel = new JLabel("col");
        colText  = new JTextField(3);
        p.add(h1);
        p.add(lineLabel);
        p.add(lineText);
        p.add(colLabel);
        p.add(colText);

        add(p);
    }

    private void createDialog()
    {
       toolsDialog   = new EditorToolsDialog(text);
    }

    protected void setListener()
    {
        cutBtn.addActionListener(new ActionListener()
        {
            public void actionPerformed(ActionEvent e)
            {
                text.cut();
            }
        });

        copyBtn.addActionListener(new ActionListener()
        {
            public void actionPerformed(ActionEvent e)
            {
                text.copy();
            }
        });

        pasteBtn.addActionListener(new ActionListener()
        {
            public void actionPerformed(ActionEvent e)
            {
                text.paste();
            }
        });

        searchBtn.addActionListener(new ActionListener()
        {
            public void actionPerformed(ActionEvent e)
            {
                openSearchDialog();
            }
        });

        macroRecBtn.addActionListener(new ActionListener()
        {
            public void actionPerformed(ActionEvent e)
            {
                macroRecording();
            }
        });

        macroPlayBtn.addActionListener(new ActionListener()
        {
            public void actionPerformed(ActionEvent e)
            {
                macroPlay();
            }
        });

        settingBtn.addActionListener(new ActionListener()
        {
            public void actionPerformed(ActionEvent e)
            {
                openSettingDialog();
           }
        });

        undoBtn.addActionListener(undoAction);
        redoBtn.addActionListener(redoAction);

        text.addFocusListener(srcFocusListener);

        text.addCaretListener(new CaretListener()
        {
           public void caretUpdate(CaretEvent e)
           {
              lineText.setText("" + getLinePos());
              colText.setText("" + getColumnPos());
           }
        });

        text.addMouseListener(new  MouseAdapter()
        {
            public void mouseClicked(MouseEvent e)
            {
               if (e.getClickCount() == 1)
               {
                   e.consume();
                   selectionBackup(e);
               }
               else if (e.getClickCount() == 2)
               {
                   e.consume();
                   doChangeSelection(e);
               }
            }
        });

        text.addKeyListener(new KeyAdapter()
        {
            public void keyPressed(KeyEvent e)
            {
                if (recording)
                    addMacroKey(e);
            }
            public void keyTyped(KeyEvent e)
            {
                if (recording)
                    addMacroKey(e);
            }
            public void keyReleased(KeyEvent e)
            {
                if (recording)
                    addMacroKey(e);
            }
        });

        lineText.addActionListener(new ActionListener()
        {
            public void actionPerformed(ActionEvent e)
            {
                goPosition();
            }
        });

        colText.addActionListener(new ActionListener()
        {
            public void actionPerformed(ActionEvent e)
            {
                goPosition();
            }
        });

        incSearchText.getDocument().addDocumentListener(new DocumentListener()
        {
            public void changedUpdate(DocumentEvent de)
            {
                toolsDialog.search.search(incSearchText.getText());
            }
            public void insertUpdate(DocumentEvent de)
            {
                toolsDialog.search.search(incSearchText.getText());
            }
            public void removeUpdate(DocumentEvent de)
            {
                toolsDialog.search.search(incSearchText.getText());
            }
        });

        incSearchText.addKeyListener(new KeyAdapter()
        {
            public void keyPressed(KeyEvent e)
            {
                if (e.getKeyCode() == KeyEvent.VK_ENTER)
                {
                    toolsDialog.search.search(incSearchText.getText(),
                                              (e.getModifiers() & KeyEvent.SHIFT_MASK) != 0);
                }
                else if (e.isAltDown() && e.getKeyCode() == KeyEvent.VK_I)
                {
                    text.requestFocus();
                }
            }
        });
    }

    public void dispose()
    {
        if (toolsDialog != null)
            toolsDialog.dispose();
        toolsDialog = null;

        toolTbl.remove(text);

        if (undoManager != null)
        {
            undoManager.dispose();
            undoManager = null;
        }

        if (keyRecords != null)
        {
            keyRecords.clear();
            keyRecords = null;
        }

        undoAction  = null;
        redoAction  = null;
        srcFocusListener =  null;
        codeParseAction = null;
        lnview   = null;
        keybinds = null;
        text     = null;
    }

    public void openSettingDialog()
    {
        if (settingDlg == null)
        {
            settingDlg = new EditSettingDialog(this);
        }
        else
        {
            settingDlg.setSourceToolBar(this);
        }
        settingDlg.setVisible(true);
    }

    public void openSearchDialog()
    {
        toolsDialog.setVisible(true);
        toolsDialog.tabs.setSelectedIndex(0);
    }

    //////////////////////////////////////////////////////

    ArrayList<KeyEvent> keyRecords = new ArrayList<KeyEvent>();
    boolean recording = false;
    void macroRecording()
    {
        if (!recording)
        {
            recording = true;
            keyRecords.clear();
            macroPlayBtn.setEnabled(false);
            macroRecBtn.setIcon(getIconImage("MacroOn.GIF"));
            macroRecBtn.setToolTipText("記録中");
        }
        else
        {
            recording = false;
            macroPlayBtn.setEnabled(true);
            macroRecBtn.setIcon(getIconImage("MacroOff.GIF"));
            macroRecBtn.setToolTipText("入力を記録");
        }
    }

    void addMacroKey(KeyEvent e)
    {
        keyRecords.add(e);
    }

    void macroPlay()
    {
        int id = 1000;
        if (keyRecords.size() == 0)
            return;

        Thread th = new Thread(new Runnable() 
        {
            public void run() { macroPlaySub(); }
        });
        th.start();
    }

    void macroPlaySub()
    {
        try
        {
            recording = false;
            macroRecBtn.setEnabled(false);
            for(KeyEvent e : keyRecords)
            {
                final KeyEvent keyEvent = new KeyEvent(text,
                                                 e.getID(), 
                                                 System.currentTimeMillis(),
                                                 e.getModifiers(), 
                                                 e.getKeyCode(),
                                                 e.getKeyChar());
                try
                {
                    SwingUtilities.invokeAndWait(new Runnable()
                    {
                        public void run()
                        {
                            text.dispatchEvent(keyEvent);
                        }
                    });
                }
                catch(InterruptedException ex){ break; }
                catch(InvocationTargetException ex){ break; }
            }
        }
        finally
        {
            macroRecBtn.setEnabled(true);
            text.requestFocus();
        }
    }

    /////////////////////////////////////////////////////

    public void resetProperties()
    {
        text.setFont(EditorConfig.font);
        lnview.resetFont(EditorConfig.font);
        text.setForeground(EditorConfig.fgColor);
        text.setBackground(EditorConfig.bgColor);
        doc.putProperty("tabSize", new Integer(EditorConfig.tabSize));

        if (doc instanceof HighLightedDocument)
        {
             HighLightedDocument h = (HighLightedDocument)doc;
             h.resetStylesToDocument();
             h.fireUpdate();
        }
        //text.repaint();
        try
        {
             toolsDialog.resetFont();
        }
        catch(Exception e)
        {
        }
    }

    /////////////////////////////////////////////////////////////////////////////

    transient private UndoStateChangeListener undoChangeListener =  new UndoStateChangeListener()
    {
        public void changed(Object mng)
        {
            undoStateChanged(mng);
        }
    };

    private UndoStateChangeListener stateChange = null;

    public void undoClear()
    {
        undoManager.clear();
    }

    public void setUndoStateListener(UndoStateChangeListener stateChange)
    {
        this.stateChange = stateChange;
    }

    private void undoStateChanged(Object manager)
    {
        undoBtn.setEnabled(undoManager.canUndo());
        redoBtn.setEnabled(undoManager.canRedo());
        
        if (stateChange != null)
            stateChange.changed(undoManager);
    }

    /////////////////////////////////////////////////////////////////////////////

    public String filePath = null;

    public void setFilePath(String fpath)
    {
        this.filePath = fpath;
    }

    ///////////////////////////////////////////////////////////////////////////

    final static String ident = "abcdefghijklmnopqrstuvwxyz_0123456789\n\'\"";

    int startSelBk = -1;
    int endSelBk = -1;

    void selectionBackup(MouseEvent ev)
    {
        int s = text.getSelectionStart();
        int e = text.getSelectionEnd();
        int ca = text.getCaretPosition();
        if (!(ca >= startSelBk && ca <= endSelBk+1))
        {
            startSelBk = -1;
            endSelBk = -1;
        }
    }

    public void doChangeSelection(MouseEvent ev)
    {
        boolean isShiftDown = ev.isShiftDown();
        int s = text.getSelectionStart();
        int e = text.getSelectionEnd();
        int ca = text.getCaretPosition();

        boolean noselect =  (s == e);
        if (ca >= startSelBk && ca <= endSelBk+1)
        {
            noselect = false;
            s = startSelBk;
            e = endSelBk;
        }

        String ss = delCR(text.getText());

        int len = ss.length();
        if (noselect || !isShiftDown)
        {
            int i = e;
            for( ; i < len;++i)
            {
                char c = Character.toLowerCase(ss.charAt(i));
                if (ident.indexOf(c) != -1)
                    break;
                else
                    e = i + 1;
            }
            for( ; i < len;++i)
            {
                char c = Character.toLowerCase(ss.charAt(i));
                if (ident.indexOf(c) == -1)
                    break;
                else
                    e = i + 1;
            }
        }

        if (noselect || isShiftDown)
        {
            int i = s;
            if (isShiftDown)
                --i;
            for( ; i >= 0; --i)
            {
                char c = Character.toLowerCase(ss.charAt(i));
                if (ident.indexOf(c) != -1)
                    break;
                else
                    s = i;
            }
            for( ; i >= 0; --i)
            {
                char c = Character.toLowerCase(ss.charAt(i));
                if (ident.indexOf(c) == -1)
                    break;
                else
                    s = i;
            }
        }

        text.select(s, e);
        startSelBk = s;
        endSelBk   = e;
    }

    ///////////////////////////////////////////////////////////////////////////

    public int getCaretPosition()
    {
        return text.getCaretPosition();
    }

    public void  setCaretPosition(int pos)
    {
        try
        {
            text.setCaretPosition(pos);
        }
        catch(Exception ex){ ex.printStackTrace(); }
    }

    public String getSelectedText()
    {
        String s = delCR(text.getSelectedText());
        return (s == null) ? "" : s;
    }

    public void replaceSelection(String s)
    {
        if (s != null)
        {
            text.replaceSelection(s);
            text.requestFocus();
        }
    }

    public void setSelection(int first,int end)
    {
        text.select(first, end);
        text.requestFocus();
    }

    public void setSelectionAll()
    {
        text.select(0, text.getText().length());
        text.requestFocus();
    }

    public void insertString(String s)
    {
        if (s != null)
        {
            try
            {
                doc.insertString(text.getCaretPosition(), s, null);
                text.requestFocus();
            }
            catch(Exception e) {}
        }
    }

    public String getText()
    {
        String s = delCR(text.getText());
        return (s == null) ? "" : s;
    }

    public void setText(String s)
    {
        text.setText(s);
        text.requestFocus();
    }

    public void remove(int start, int end)
    {
        text.select(start, end);
        text.replaceSelection("");
        text.requestFocus();
    }

    public void replace(int start, int end, String s)
    {
        text.select(start, end);
        text.replaceSelection(s);
        text.requestFocus();
    }

    //////////////////////////////////////

    public int getLinePos()
    {
        String s = delCR(text.getText());
        int pos  = text.getCaretPosition();
        int cnt  = 0;

        for (int i = 0; i < pos ; ++i)
        {
            if (s.charAt(i) == '\n')
               ++cnt;
        }
        return cnt + 1;
    }

    public int getColumnPos()
    {
        String s = delCR(text.getText());
        int pos  = text.getCaretPosition();
        int cnt  = 0;

        for (int i = pos - 1; i >= 0 ; --i, ++cnt)
        {
            if (s.charAt(i) == '\n')
               break;
        }
        return cnt + 1;
    }

    public void goPosition()
    {
        String s = delCR(text.getText());

        try
        {
            int li = Integer.parseInt(lineText.getText());
            int co = Integer.parseInt(colText.getText());

            int st = getLineStartPos(li);
            st +=  co - 1;

            if (st < 0)
               st = 0;

            if (st >= s.length())
               st = s.length();

            text.setCaretPosition(st);
       }
       catch(Exception e)
       {
       }
       finally
       {
          text.requestFocus();
       }
    }

    public int getLineStartPos(int line)
    {
        if (line <= 1)
           return 0;

        String s = delCR(text.getText());
        int cnt = 1;

        int i = 0;
        for (; i < s.length() && cnt < line; ++i)
        {
            if (s.charAt(i) == '\n')
                ++cnt;
        }
        return i;
    }

    public int getLineSize()
    {
        String s = delCR(text.getText());
        StringTokenizer tok = new StringTokenizer(s, "\n");
        return tok.countTokens();
    }

    public String getLine(int line)
    {
        if (line < 0 )
           return "";

        String s = delCR(text.getText());
        return s.substring(getLineStartPos(line), getLineStartPos(line+1));
    }

    public void setLine(int line, String s)
    {
        if (line < 0  || s == null)
           return ;

        String r = delCR(s);
        text.select(getLineStartPos(line), getLineStartPos(line+1));
        text.replaceSelection(r);
    }

    public static String delCR(String s)
    {
        if (s == null)
            return "";

        StringBuffer b = new StringBuffer(s.length());
        for(int i=0; i< s.length(); ++i)
        {
            char a = s.charAt(i);
            if (a !='\r')
               b.append(a);
        }
        return b.toString();
    }

    public void undo()
    {
        undoManager.execUndo();
        text.requestFocus();
    }

    public void redo()
    {
        undoManager.execRedo();
        text.requestFocus();
    }

    public void searchNext()
    {
        String  sel = delCR(text.getSelectedText());
        toolsDialog.search.setReverse(false);
        toolsDialog.search.search(sel);
    }

    public void searchPrev()
    {
        String  sel = delCR(text.getSelectedText());
        toolsDialog.search.setReverse(true);
        toolsDialog.search.search(sel);
    }

    public void transNext()
    {
        if (toolsDialog.search.isVisible())
        {
            toolsDialog.search.allOff();
            toolsDialog.search.translate();
        }
        else
        {
            toolsDialog.search.setVisible(true);
            String  sel = delCR(text.getSelectedText());
            toolsDialog.search.allOff();
            toolsDialog.search.setTranseFocus(sel);
        }
    }

    public void jumpLine()
    {
        colText.setText("1");
        lineText.selectAll();
        lineText.requestFocus();
    }

    public void focusedIncText()
    {
        incSearchText.requestFocus();
    }

    ///////////////////////////////////////////////////////////

    public void searchBlock()
    {
        String  sel = delCR(text.getSelectedText());
        if (sel == null || sel.isEmpty())
        {
            String s = delCR(text.getText());
            int spos =  text.getCaretPosition();
            sel = "" + s.charAt(spos);

            if (sel == null || sel.isEmpty())
               return;

            if (sel.equals("(") || sel.equals(")")
             || sel.equals("{") || sel.equals("}")
             || sel.equals("[") || sel.equals("]"))
            {
                setSelection(spos, spos+1);
            }
        }

        boolean rev = false;
        char st;
        char ed;
        if (sel.equals("{"))
        {
            st = '{';
            ed = '}';
        }
        else if (sel.equals("("))
        {
            st = '(';
            ed = ')';
        }
        else if (sel.equals("["))
        {
            st = '[';
            ed = ']';
        }
        else if (sel.equals("}"))
        {
            rev = true;
            st = '}';
            ed = '{';
        }
        else if (sel.equals(")"))
        {
            rev = true;
            st = ')';
            ed = '(';
        }
        else if (sel.equals("]"))
        {
            rev = true;
            st = ']';
            ed = '[';
        }
        else
            return;

        if (!rev)
            seekEndBlock(st, ed);
        else
            seekStartBlock(st, ed);
    }

    private void seekEndBlock(char st, char ed)
    {
        String s   = delCR(text.getText());
        int pos    = text.getCaretPosition();
        int len    = s.length();
        int level  = 0;

        for (int i = pos; i < len ; ++i)
        {
            char c = s.charAt(i);

            if (c == st)
               level++;

            if (c == ed)
            {
               if (level == 0)
               {
                   setSelection(i, i + 1);
                   return;
               }
               level--;
            }
        }
    }

    private void seekStartBlock(char st, char ed)
    {
        String s   = delCR(text.getText());
        int pos    = text.getCaretPosition();
        int level  = 0;

        for (int i = pos - 2; i >= 0 ; --i)
        {
            char c = s.charAt(i);
            if (c == st)
               level++;

            if (c == ed)
            {
               if (level == 0)
               {
                   setSelection(i, i + 1);
                   return;
               }
               level--;
            }
        }
    }

    ///////////////////////////////////////////////////////////

    public void tabInput()
    {
        String  sel = delCR(text.getSelectedText());
        if (sel == null || sel.isEmpty())
        {
             if (EditorConfig.keyTab2space)
             {
                 int tabsize = EditorConfig.tabSize;
                 int col = getColumnPos();
                 int n = tabsize - ((col-1) % tabsize);

                 StringBuilder sb = new StringBuilder();
                 for(int i = 0; i < n; ++i)
                    sb.append(' ');
                 insertString(sb.toString());
             }
             else
             {
                 insertString("\t");
             }
        }
        else
        {
            indentAddTab(sel);
        }
    }

    public void indentTab(boolean isAdd)
    {
       if (isAdd)
          indentAddTab(null);
       else
          indentDelTab();
    }

    public void indentAddTab(String  sel)
    {
        StringBuilder spc = new StringBuilder();
        int tabsize =  EditorConfig.tabSize;
        for(int i = 0; i < tabsize; ++i)
            spc.append(' ');

        if (sel == null)
            sel = delCR(text.getSelectedText());
        if (sel != null && !sel.isEmpty())
        {
            int start = text.getSelectionStart();

            String [] tok = sel.split("\n");
            StringBuilder sb = new StringBuilder();
            for(String s : tok)
            {
                if (s.length() != 0)
                {
                    if (EditorConfig.keyTab2space)
                       sb.append(spc.toString());
                    else
                       sb.append("\t");
                }
                sb.append(s);
                sb.append("\n");
            }
            String chg = sb.toString();
            text.replaceSelection(chg);
            text.select(start, start + chg.length());
        }
    }

    public void indentDelTab()
    {
        StringBuilder spc = new StringBuilder();
        int tabsize =  EditorConfig.tabSize;
        for(int i = 0; i < tabsize; ++i)
            spc.append(' ');

        String  sel = delCR(text.getSelectedText());
        if (sel != null && !sel.isEmpty())
        {
            int start = text.getSelectionStart();
            String [] tok = sel.split("\n");
            StringBuilder sb = new StringBuilder();
            for(String s : tok)
            {
                if (s.length() > 0 && s.charAt(0) == '\t')
                    sb.append(s.substring(1));
                else if (s.startsWith(spc.toString()))
                    sb.append(s.substring(spc.length()));
                else
                    sb.append(s);

               sb.append("\n");
            }
            String chg = sb.toString();
            text.replaceSelection(chg);
            text.select(start, start + chg.length());
        }
     }

    ///////////////////////////////////////////////////////////

    public void commentOut(boolean on)
    {
       if (on)
          commentOutSet();
       else
          commentOutReset();
    }

    public void commentOutSet()
    {
        lineSelect();

        int start =  text.getSelectionStart();
        int end   =  text.getSelectionEnd();
        try
        {
            doc.insertString(end,   "\n*/\n", null);
            doc.insertString(start, "/*\n", null);
            text.select(start, end + 7);
        }
        catch(Exception ex) { ex.printStackTrace(); }
    }

    public void commentOutReset()
    {
        lineSelect();
        String  sel = delCR(text.getSelectedText());
        int start =  text.getSelectionStart();

        int flen = 0;
        int llen = 0;
        int f = sel.indexOf("/*");
        int l = sel.lastIndexOf("*/");
        if (f < 0 && l < 0)
           return;

         flen = llen = 2;
         int f2 = sel.indexOf("/*\n");
         if (f2 >= 0)
         {
             f = f2;
             flen = 3;
         }
         int l2 = sel.lastIndexOf("\n*/\n");
         if (l2 >= 0)
         {
             l  = l2;
             llen = 4;
         }

         try
         {
            doc.remove(start + l, llen);
            doc.remove(start + f, flen);
         }
         catch(Exception ex) { ex.printStackTrace(); }
    }

    ///////////////////////////////////////////////////////////

    public void lineEndDel()
    {
        String s = delCR(text.getText());
        int spos =  getCaretPosition();
        int i = spos -1;
        for (;; ++i)
        {
            if (s.charAt(i) == '\n')
            {
                i++;
                break;
            }
        }
        remove(spos, i-1);
    }

    public void lineSelect()
    {
        String s = delCR(text.getText());
        int spos =  text.getSelectionStart();
        if (spos >= 0 )
        {
             for (int i = spos - 1; i >= 0 ; --i, --spos)
             {
                 if (s.charAt(i) == '\n')
                 {
                     i++;
                     break;
                 }
             }

             int pos  = getLinePos();
             int epos = getLineStartPos(pos);
             for (int i = epos ; i < s.length() ; ++i, ++epos)
             {
                 if (s.charAt(i) == '\n')
                 {
                     break;
                 }
             }

             setSelection(spos, epos);
        }
        else
        {
             int pos  = getLinePos();
             spos = getLineStartPos(pos);
             int epos = spos + 1;

             for (int i = spos + 1 ; i < s.length() ; ++i, ++epos)
             {
                 if (s.charAt(i) == '\n')
                 {
                     break;
                 }
             }

            setSelection(spos, epos);
        }
    }

    ///////////////////////////////////////////////////////////

    public void gotoLine(int line, int col)
    {
        if (line < 1)
        {
           setCaretPosition(0);
           return;
        }

        String s = delCR(text.getText());

        if (line > getLineSize())
        {
            setCaretPosition(s.length());
            return;
        }

        col--;
        if (col < 0)
           col = 0;

        String ll = getLine(line);

        if (col > ll.length() - 1)
           col = ll.length() - 1;

        setCaretPosition(getLineStartPos(line) + col);
    }

    ///////////////////////////////////////////////////////////

    public void refreshEditor()
    {
        textValidate();
    }

    ///////////////////////////////////////////////////////

    private void keyBinding()
    {
        if (keybinds == null)
        {
            keybinds = new KeyBindings(this);
        }
        keybinds.keyBind();
    }

    public void resetKeyBinding()
    {
        keyBinding();
    }

    ///////////////////////////////////////////////////////////////////////////

    class UndoAction extends AbstractAction
    {
        public UndoAction()
        {
            super("アンドゥ");
        }

        public void actionPerformed(ActionEvent e)
        {
            try
            {
                undoManager.execUndo();
                text.requestFocus();
            }
            catch(CannotUndoException ex)
            {
            }
        }
    }

    class RedoAction extends AbstractAction
    {
        public RedoAction()
        {
            super("リドゥ");
        }

        public void actionPerformed(ActionEvent e)
        {
            try
            {
                undoManager.execRedo();
                text.requestFocus();
            }
            catch(CannotRedoException ex)
            {
            }
        }
    }

    class SourceFocusListener extends  FocusAdapter
    {
        public void focusGained(FocusEvent e)
        {
        }
    };

    ///////////////////////////////////////////////////////////////////////////

    public void initText(final String s)
    {
        Cursor old = getCursor();
        text.setCursor(Cursor.getPredefinedCursor(Cursor.WAIT_CURSOR));

        try
        {
            setText(s);
            return ;
        }
        finally
        {
           text.setCursor(old);
        }
    }

    public void textInit()
    {
        setText("");
    }

    public void textValidate()
    {
        text.validate();
        text.repaint();
    }
}

