//============================================================================
//
//    Java Text Editor library
//
//    TextUndoManager.java
//
//  Copyright (c) 2013 Fumiaki Hosaka.
//
//  Permission to copy, use, modify, sell and distribute this software
//  is granted provided this copyright notice appears in all copies.
//  This software is provided "as is" without express or implied
//  warranty, and with no claim as to its suitability for any purpose.
//============================================================================
package text.editor;

import java.awt.event.*;

import javax.swing.*;
import javax.swing.text.*;

import java.util.*;

public class TextUndoManager
{
    Stack<TextUndoCommand>  undo_stack = new Stack<TextUndoCommand> ();

    Stack<TextUndoCommand>  redo_stack = new Stack<TextUndoCommand> ();

    private UndoStateChangeListener stateChange = null;
    boolean undoing = false;
    boolean redoing = false;

    JTextComponent text = null;
    Document doc = null;

    public TextUndoManager(JTextComponent text, Document doc, UndoStateChangeListener stateChange) 
    {
        this.text = text;
        this.doc = doc;
        this.stateChange = stateChange;
    }

    public void clear()
    {
        undo_stack.clear();
        redo_stack.clear();
        if (stateChange != null)
            stateChange.changed(this);
    }

    public void dispose()
    {
        if (undo_stack != null)
        {
            undo_stack.clear();
            redo_stack.clear();
            undo_stack = null;
            redo_stack = null;
            stateChange = null;
        }
    }

    public boolean canUndo()
    {
        return (undo_stack.size() > 0);
    }

    public boolean canRedo()
    {
        return (redo_stack.size() > 0);
    }

    public synchronized boolean isDoing()
    {
        return (undoing || redoing);
    }

    public synchronized  void execUndo() 
    {
        if (undoing || redoing)
            return; 

        try
        {
            undoing = true;
            execCommand(undo_stack);
        }
        finally
        {
            undoing = false;
        }
    }

    public synchronized  void execRedo() 
    {
        if (undoing || redoing)
            return; 

        try
        {
            redoing = true;
            execCommand(redo_stack);
        }
        finally
        {
            redoing = false;
        }
    }

    private  void execCommand(Stack<TextUndoCommand> stack)  
    {
        if (stack.empty())
            return;

        TextUndoCommand c = stack.pop();

        try
        {
            if (c.type == TextUndoCommand.INSERT)
            {
                text.setCaretPosition(c.offs);
                doc.insertString(c.offs, c.str, c.att);
            } 
            else
            {
                text.setCaretPosition(c.offs);
                doc.remove(c.offs, c.len);
            }
       }
       catch(Exception ex){}
    }

    public synchronized void push(TextUndoCommand cmd) 
    {
        if (!undoing)
        {
            undo_stack.push(cmd);
            if (stateChange != null)
                stateChange.changed(this);
        }
        else if (!redoing)
        {
            redo_stack.push(cmd);
            if (stateChange != null)
                stateChange.changed(this);
        }
    }
}

