//============================================================================
//
//    Java Text Editor library
//
//    CustomParagraphView.java
//
//  Copyright (c) 2013 Fumiaki Hosaka.
//
//  Permission to copy, use, modify, sell and distribute this software
//  is granted provided this copyright notice appears in all copies.
//  This software is provided "as is" without express or implied
//  warranty, and with no claim as to its suitability for any purpose.
//============================================================================
package text.editor;

import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import javax.swing.plaf.*;
import javax.swing.text.*;

class CustomParagraphView extends ParagraphView 
{
    private static Color pc = new Color(130, 140, 120);

    static public void setGraphColor(Color c)
    {
        pc = c;
    }

    boolean canEdit;
    public CustomParagraphView(Element elem, boolean canEdit) 
    {
        super(elem);
        this.canEdit = canEdit;
    }

    protected SizeRequirements calculateMinorAxisRequirements(int axis, SizeRequirements r) 
    {
        SizeRequirements req = super.calculateMinorAxisRequirements(axis, r);
        req.minimum = req.preferred;
        return req;
    }

    public int getFlowSpan(int index) 
    {
        return Integer.MAX_VALUE;
    }

    public void paint(Graphics g, Shape allocation)
    {
        super.paint(g, allocation);
        if (!canEdit)
            return;

        paintCustomParagraph(g, allocation);
    }

    private void paintCustomParagraph(Graphics g, Shape a)
    {
        try
        {
            Shape paragraph = modelToView(getEndOffset(), a, Position.Bias.Backward);
            Rectangle r = (paragraph==null)?a.getBounds():paragraph.getBounds();
            int x = r.x;
            int y = r.y;
            int h = r.height;
            FontMetrics fontMetrics = g.getFontMetrics();
            int w = fontMetrics.stringWidth(" ");
            Color old = g.getColor();
            g.setColor(pc);

            g.drawLine(x+1,  y+h/2,   x+w,  y+h/2);
            g.drawLine(x+w,  y+h/4,   x+w,  y+h/2);

            int px = x+1;
            int py = y+h/2;
            int px1=px,py1=py,px2=px,py2=py;
            for(int i = 1; i <= w/3; ++i)
            {
                px1 = px+i;
                py1 = py-i;
                px2 = px+i;
                py2 = py+i;
            }
            g.drawLine(px,  py,  px1,  py1);
            g.drawLine(px,  py,  px2,  py2);
            g.setColor(old);
        }
        catch(Exception e) { e.printStackTrace(); }
    }
}

class WhitespaceLabelView extends LabelView
{
    static public void setGraphColor(Color c)
    {
        pc = c;
    }

    private static Color pc = new Color(130, 140, 120);
    private static final BasicStroke dashed = new BasicStroke(1.0f, BasicStroke.CAP_BUTT, 
                                                BasicStroke.JOIN_MITER, 10.0f, new float[] {1.0f}, 0.0f);

    boolean canEdit;

    public WhitespaceLabelView(Element elem, boolean canEdit)
    {
        super(elem);
        this.canEdit = canEdit;
    }

    public void paint(Graphics g, Shape a)
    {
        super.paint(g,a);

        if (!canEdit)
            return;

        Graphics2D g2 = (Graphics2D)g;
        Stroke stroke = g2.getStroke();
        Rectangle alloc = (a instanceof Rectangle) ? (Rectangle)a : a.getBounds();
        FontMetrics fontMetrics = g.getFontMetrics();
        int spaceWidth = fontMetrics.stringWidth(" ");
        int sumOfTabs  = 0;
        String text = getText(getStartOffset(),getEndOffset()).toString();
        for(int i=0;i<text.length();i++) 
        {
            String s = text.substring(i,i+1);
            int previousStringWidth = fontMetrics.stringWidth(text.substring(0,i)) + sumOfTabs;
            int sx = alloc.x+previousStringWidth;
            int sy = alloc.y+alloc.height-fontMetrics.getDescent();
            if(" ".equals(s))
            {
                g2.setStroke(dashed);
                g2.setPaint(pc);
                g2.drawLine(sx+1, sy-1, sx+spaceWidth-2, sy-1);
                g2.drawLine(sx+2,   sy, sx+spaceWidth-2, sy);
            }
            else if("\t".equals(s))
            {
                int tabWidth = (int)getTabExpander().nextTabStop((float)sx, i)-sx;
                g2.setColor(pc);
                g2.drawLine(sx+2, sy-0, sx+2+2, sy-0);
                g2.drawLine(sx+2, sy-1, sx+2+1, sy-1);
                g2.drawLine(sx+2, sy-2, sx+2+0, sy-2);
                g2.setStroke(dashed);
                g2.drawLine(sx+2, sy, sx+tabWidth-2, sy);
                sumOfTabs+=tabWidth;
            }
            g2.setStroke(stroke);
        }
    }
}

class CustomViewFactory implements ViewFactory
{
    boolean canEdit;
    public CustomViewFactory(boolean canEdit)
    {
        super();
        this.canEdit = canEdit;
    }

    public View create(Element elem)
    {
        String kind = elem.getName();
        if (kind != null) 
        {
            if(kind.equals(AbstractDocument.ContentElementName))
            {
                return new WhitespaceLabelView(elem, canEdit);
            }
            else if (kind.equals(AbstractDocument.ParagraphElementName))
            {
                return new CustomParagraphView(elem, canEdit);
            }
            else if (kind.equals(AbstractDocument.SectionElementName))
            {
                return new BoxView(elem, View.Y_AXIS);
            }
            else if (kind.equals(StyleConstants.ComponentElementName))
            {
                return new ComponentView(elem);
            }
            else if (kind.equals(StyleConstants.IconElementName))
            {
                return new IconView(elem);
            }
        }
        return new LabelView(elem);
    }
}

